@@ -11,6 +11,7 @@ import gevent
 
 
 API_VERSION = 'v1'
+RETRY_AFTER = 30 # seconds
 
 def _check_for_error(r):
     if not r.ok:
@@ -193,21 +194,7 @@ class _Datasource(_Base):
         if model_path is not None:
             self.trainers[key]['model_path'] = model_path
 
-        for p in self.dfl.project_participants:
-            project_key = p['project_key']
-            datasource_key = p['datasource_key']
-            threads_key = get_threads_key(project_key, datasource_key)
-            '''
-            The first time attach_datasource() we might find 1 previous project for that ds1.
-            So project_participants = [item1]
-            The next time it is called it might also find 1 previous project for ds2
-            So project_participants = [item1, item2]
-            We only want create a thread for items that we haven't done so already
-            '''
-            if threads_key in self.dfl.on_round_threads:
-                continue
-            info = self._make_request('GET', f'/projects/{project_key}')
-            self.dfl.on_round_threads[threads_key] = gevent.spawn(self.dfl.train_and_test_callback, datasource_key, info)
+        self.dfl.initiate_project_participants(should_spawn_threads=True)
 
 
 class DynamoFL(_Base):
@@ -239,6 +226,7 @@ class DynamoFL(_Base):
         j = json.loads(res)
         if j['event'] == 'client-info':
             self.instance_id = j['data']['id']
+            self.initiate_project_participants(should_fetch_bridges=True, should_spawn_threads=True)
 
         if j['event'] == 'new-project':
             project_key = j['data']['projectKey']
@@ -284,6 +272,12 @@ class DynamoFL(_Base):
                 if threads_key in self.on_round_threads:
                     print('Federation error occured:\n  ' + j['data']['errorMessage'])
 
+
+    def kill_threads(self):
+        for threads_key in self.on_round_threads:
+            self.on_round_threads[threads_key].kill()
+        self.on_round_threads = {}
+
     def connect_to_ws(self):
         t = threading.Thread(target=self.ws.run_forever)
         t.daemon = False
@@ -291,8 +285,9 @@ class DynamoFL(_Base):
 
     def _on_close(self, ws, close_status_code, close_msg):
         print('Connection closed')
-        print("Trying again to connect in 30 seconds...")
-        r = threading.Timer(30, self.connect_to_ws)
+        self.kill_threads()
+        print(f'Trying to reestablish connection...')
+        r = threading.Timer(RETRY_AFTER, self.connect_to_ws)
         r.start()
 
     def _on_error(self, ws, error):
@@ -363,6 +358,35 @@ class DynamoFL(_Base):
             print('Done.')
             print()
 
+    def initiate_project_participants(self, should_fetch_bridges=False, should_spawn_threads=False):
+        self.project_participants = []
+        if should_fetch_bridges:
+            for ds_key in self.datasources:
+                j = self._make_request('GET', '/bridges', params={'datasourceKey': ds_key})
+                for i in j['data']:
+                    self.project_participants.append({
+                        'project_key': i['projectKey'],
+                        'datasource_key': i['datasourceKey'],
+                        'trainer_key': i['trainerKey'],
+                        'hyper_param_values': i['hyperParamValues']
+                    })
+        if should_spawn_threads:
+            for p in self.project_participants:
+                project_key = p['project_key']
+                datasource_key = p['datasource_key']
+                threads_key = get_threads_key(project_key, datasource_key)
+                '''
+                The first time attach_datasource() we might find 1 previous project for that ds1.
+                So project_participants = [item1]
+                The next time it is called it might also find 1 previous project for ds2
+                So project_participants = [item1, item2]
+                We only want create a thread for items that we haven't done so already
+                '''
+                if threads_key in self.dfl.on_round_threads:
+                    continue
+                info = self._make_request('GET', f'/projects/{project_key}')
+                self.on_round_threads[threads_key] = gevent.spawn(self.train_and_test_callback, datasource_key, info)    
+                
 
     # creates a new datasource in the api
     def attach_datasource(self, key, name=None, metadata=None):
@@ -385,14 +409,7 @@ class DynamoFL(_Base):
             self._make_request('POST', '/datasources', params=params)
 
         # get current projects of newly attached datasource
-        j = self._make_request('GET', '/bridges', params={'datasourceKey': key})
-        for i in j['data']:
-            self.project_participants.append({
-                'project_key': i['projectKey'],
-                'datasource_key': i['datasourceKey'],
-                'trainer_key': i['trainerKey'],
-                'hyper_param_values': i['hyperParamValues']
-            })
+        self.initiate_project_participants(should_fetch_bridges=True)
 
         ds = _Datasource(self, key)
         self.datasources[key] = ds
